@@ -33,12 +33,14 @@ esac
 # getloadavg: ditto: https://builds.sr.ht/~alip/job/1395813 TODO: ditto!
 # localename-environ: config fails: configure.ac:41: error: possibly undefined macro: gl_LOCALE_MODULE_INDICATOR
 # poll: racy test, fails randomly, passes most of the time: https://builds.sr.ht/~alip/job/1454355#task-test64
-SKIP_PATTERN='(backupfile|bison|copy-file|datetime|exception|exclude|gettext|link-warning|localename-environ|mbsstr|nan|poll|printf|trapping|vma-prot|wctype|windows|^(array|java|uni)|-ieee$|(^(execinfo|getloadavg|gnumakefile|havelib|physmem|timevar)$))'
+# quotearg: https://builds.sr.ht/~alip/job/1533212#task-test32 TODO: Figure out why!
+SKIP_PATTERN='(backupfile|bison|copy-file|datetime|exception|exclude|gettext|link-warning|localename-environ|mbsstr|nan|poll|printf|quotearg|trapping|vma-prot|wctype|windows|^(array|java|uni)|-ieee$|(^(execinfo|getloadavg|gnumakefile|havelib|physmem|timevar)$))'
 
 # Tests that have failed in the past.
 # copy-file fails to compile on i586 so we skip it.
 # linkat: https://builds.sr.ht/query/log/1402429/test32/log
 # mkdir: https://builds.sr.ht/~alip/job/1286254
+# openat-safer: https://builds.sr.ht/~alip/job/1522333
 # posix_openpt: https://builds.sr.ht/~alip/job/126488
 # readdir: no known fail but best to check everytime.
 # rename: https://builds.sr.ht/~alip/job/1286933
@@ -47,6 +49,7 @@ SKIP_PATTERN='(backupfile|bison|copy-file|datetime|exception|exclude|gettext|lin
 FAIL_HISTORY=(
     linkat
     mkdir
+    openat-safer
     posix_openpt
     readdir
     rename
@@ -65,9 +68,9 @@ ulimit -c unlimited
 # Force TTY output.
 export SYD_FORCE_TTY=YesPlease
 
-# Timeout is 10 minutes per-test,
+# Timeout is 15 minutes per-test,
 # unless otherwise specified.
-SYD_TEST_TIMEOUT=${SYD_TEST_TIMEOUT:-10m}
+SYD_TEST_TIMEOUT=${SYD_TEST_TIMEOUT:-15m}
 
 export SYD_LOG=${SYD_LOG:-info}
 SYD="${CARGO_BIN_EXE_syd:-syd}"
