@@ -35,7 +35,7 @@ use nix::{
 use once_cell::sync::Lazy;
 use serde::{ser::SerializeMap, Serialize, Serializer};
 
-use crate::{config::*, fs::FileType, IoctlRequest, XPath};
+use crate::{config::*, fs::FileType, XPath};
 
 #[repr(C)]
 pub(crate) struct TimeSpec64 {
@@ -1415,8 +1415,8 @@ pub fn epoll_ctl_mod_safe<E: AsFd>(
     .map(drop)
 }
 
-const EPIOCSPARAMS: IoctlRequest = 0x40088a01;
-const EPIOCGPARAMS: IoctlRequest = 0x80088a02u32 as IoctlRequest;
+const EPIOCSPARAMS: u64 = 0x40088a01;
+const EPIOCGPARAMS: u64 = 0x80088a02;
 
 /// Epoll parameters
 #[repr(C)]
@@ -1469,7 +1469,15 @@ impl Serialize for EpollParams {
 /// Requires Linux>=6.9.
 pub fn epoll_set_params<Fd: AsFd>(fd: Fd, params: &EpollParams) -> Result<(), Errno> {
     // SAFETY: In libc we trust.
-    Errno::result(unsafe { libc::ioctl(fd.as_fd().as_raw_fd(), EPIOCSPARAMS, params) }).map(drop)
+    Errno::result(unsafe {
+        libc::syscall(
+            libc::SYS_ioctl,
+            fd.as_fd().as_raw_fd(),
+            EPIOCSPARAMS,
+            params,
+        )
+    })
+    .map(drop)
 }
 
 /// Get epoll parameters for the given epoll file descriptor.
@@ -1479,7 +1487,12 @@ pub fn epoll_get_params<Fd: AsFd>(fd: Fd) -> Result<EpollParams, Errno> {
 
     // SAFETY: In libc we trust.
     Errno::result(unsafe {
-        libc::ioctl(fd.as_fd().as_raw_fd(), EPIOCGPARAMS, params.as_mut_ptr())
+        libc::syscall(
+            libc::SYS_ioctl,
+            fd.as_fd().as_raw_fd(),
+            EPIOCGPARAMS,
+            params.as_mut_ptr(),
+        )
     })?;
 
     // SAFETY: ioctl returned success.
