@@ -64,6 +64,11 @@ pub(crate) fn handle_accept(
         None
     };
 
+    let sandbox = request.get_sandbox();
+    let force_cloexec = sandbox.force_cloexec();
+    let force_rand_fd = sandbox.force_rand_fd();
+    drop(sandbox); // release read-lock.
+
     let mut flags = if op == 0x12 {
         // accept4
         SockFlag::from_bits(args[3].try_into().or(Err(Errno::EINVAL))?).ok_or(Errno::EINVAL)?
@@ -71,7 +76,7 @@ pub(crate) fn handle_accept(
         // accept
         SockFlag::empty()
     };
-    let cloexec = flags.contains(SockFlag::SOCK_CLOEXEC);
+    let cloexec = force_cloexec || flags.contains(SockFlag::SOCK_CLOEXEC);
     flags.insert(SockFlag::SOCK_CLOEXEC);
 
     // SAFETY: Record blocking call so it can get invalidated.
@@ -135,7 +140,7 @@ pub(crate) fn handle_accept(
     }
 
     // Send the fd and return.
-    request.send_fd(fd, cloexec)
+    request.send_fd(fd, cloexec, force_rand_fd)
 }
 
 fn do_accept4(
