@@ -1,7 +1,12 @@
+// SPDX-License-Identifier: Apache-2.0 OR MIT
+
 use std::{
     io::Error,
     mem::size_of_val,
-    os::unix::io::{IntoRawFd, RawFd},
+    os::{
+        fd::{AsRawFd, FromRawFd, OwnedFd},
+        unix::io::{IntoRawFd, RawFd},
+    },
 };
 
 use libc::close;
@@ -9,9 +14,9 @@ use libc::close;
 #[cfg(test)]
 use crate::landlock::*;
 use crate::landlock::{
-    compat::private::OptionCompatLevelMut, uapi, Access, AccessFs, AccessNet, AddRuleError,
-    AddRulesError, CompatLevel, CompatState, Compatibility, Compatible, CreateRulesetError,
-    RestrictSelfError, RulesetError, ScopeFlag, TryCompat,
+    access::PrivateAccess, compat::private::OptionCompatLevelMut, uapi, Access, AccessFs,
+    AccessNet, AddRuleError, AddRulesError, CompatLevel, CompatState, Compatibility, Compatible,
+    CreateRulesetError, RestrictSelfError, RulesetError, Scope, TryCompat,
 };
 
 // Public interface without methods and which is impossible to implement outside this crate.
@@ -176,10 +181,10 @@ fn support_no_new_privs() -> bool {
 pub struct Ruleset {
     pub(crate) requested_handled_fs: AccessFs,
     pub(crate) requested_handled_net: AccessNet,
-    pub(crate) requested_scoped: ScopeFlag,
+    pub(crate) requested_scoped: Scope,
     pub(crate) actual_handled_fs: AccessFs,
     pub(crate) actual_handled_net: AccessNet,
-    pub(crate) actual_scoped: ScopeFlag,
+    pub(crate) actual_scoped: Scope,
     pub(crate) compat: Compatibility,
 }
 
@@ -263,6 +268,7 @@ impl Ruleset {
                     assert!(
                         !self.requested_handled_fs.is_empty()
                             || !self.requested_handled_net.is_empty()
+                            || !self.requested_scoped.is_empty()
                     );
 
                     // CompatState::No should be handled as CompatState::Dummy because it is not
@@ -272,7 +278,7 @@ impl Ruleset {
                         CompatLevel::HardRequirement => {
                             Err(CreateRulesetError::MissingHandledAccess)
                         }
-                        _ => Ok(RulesetCreated::new(self, -1)),
+                        _ => Ok(RulesetCreated::new(self, None)),
                     }
                 }
                 CompatState::Full | CompatState::Partial => {
@@ -290,7 +296,10 @@ impl Ruleset {
                         scoped: self.actual_scoped.bits(),
                     };
                     match unsafe { uapi::landlock_create_ruleset(&attr, size_of_val(&attr), 0) } {
-                        fd if fd >= 0 => Ok(RulesetCreated::new(self, fd)),
+                        fd if fd >= 0 => Ok(RulesetCreated::new(
+                            self,
+                            Some(unsafe { OwnedFd::from_raw_fd(fd) }),
+                        )),
                         _ => Err(CreateRulesetError::CreateRulesetCall {
                             source: Error::last_os_error(),
                         }),
@@ -353,6 +362,20 @@ pub trait RulesetAttr: Sized + AsMut<Ruleset> + Compatible {
         T::ruleset_handle_access(self.as_mut(), access)?;
         Ok(self)
     }
+
+    /// Attempts to add a set of scopes that will be supported by this ruleset.
+    /// Consecutive calls to `scope()` will be interpreted as logical ORs
+    /// with the previous scopes.
+    ///
+    /// On error, returns a wrapped [`ScopeError`](crate::ScopeError).
+    /// E.g., `RulesetError::Scope(ScopeError)`
+    fn scope<T>(mut self, scope: T) -> Result<Self, RulesetError>
+    where
+        T: Into<Scope>,
+    {
+        Scope::ruleset_handle_access(self.as_mut(), scope.into())?;
+        Ok(self)
+    }
 }
 
 impl RulesetAttr for Ruleset {}
@@ -385,13 +408,21 @@ fn ruleset_attr() {
 
 #[test]
 fn ruleset_created_handle_access_fs() {
+    let access = make_bitflags!(AccessFs::{Execute | ReadDir});
+
     // Tests AccessFs::ruleset_handle_access()
+    let ruleset = Ruleset::from(ABI::V1).handle_access(access).unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access);
+    assert_eq!(ruleset.actual_handled_fs, access);
+
+    // Tests composition (binary OR) of handled accesses.
     let ruleset = Ruleset::from(ABI::V1)
         .handle_access(AccessFs::Execute)
         .unwrap()
         .handle_access(AccessFs::ReadDir)
+        .unwrap()
+        .handle_access(AccessFs::Execute)
         .unwrap();
-    let access = make_bitflags!(AccessFs::{Execute | ReadDir});
     assert_eq!(ruleset.requested_handled_fs, access);
     assert_eq!(ruleset.actual_handled_fs, access);
 
@@ -437,6 +468,82 @@ fn ruleset_created_handle_access_net_tcp() {
     ));
 }
 
+#[test]
+fn ruleset_created_scope() {
+    let scopes = make_bitflags!(Scope::{AbstractUnixSocket | Signal});
+
+    // Tests Ruleset::scope() with ABI that doesn't support scopes.
+    let ruleset = Ruleset::from(ABI::V5).scope(scopes).unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, Scope::EMPTY);
+
+    // Tests Ruleset::scope() with ABI that supports scopes.
+    let ruleset = Ruleset::from(ABI::V6).scope(scopes).unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+
+    // Tests composition (binary OR) of scopes.
+    let ruleset = Ruleset::from(ABI::V6)
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap()
+        .scope(Scope::Signal)
+        .unwrap()
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap();
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+
+    // Tests that only the required scopes are reported as incompatible:
+    // scope should not contain Scope::AbstractUnixSocket.
+    assert!(matches!(Ruleset::from(ABI::Unsupported)
+        .scope(Scope::AbstractUnixSocket)
+        .unwrap()
+        .set_compatibility(CompatLevel::HardRequirement)
+        .scope(Scope::Signal)
+        .unwrap_err(),
+        RulesetError::HandleAccesses(HandleAccessesError::Scope(HandleAccessError::Compat(
+            CompatError::Access(AccessError::Incompatible { access })
+        ))) if access == Scope::Signal
+    ));
+}
+
+#[test]
+fn ruleset_created_fs_net_scope() {
+    let access_fs = make_bitflags!(AccessFs::{Execute | ReadDir});
+    let access_net = make_bitflags!(AccessNet::{BindTcp | ConnectTcp});
+    let scopes = make_bitflags!(Scope::{AbstractUnixSocket | Signal});
+
+    // Tests composition (binary OR) of handled accesses.
+    let ruleset = Ruleset::from(ABI::V5)
+        .handle_access(access_fs)
+        .unwrap()
+        .scope(scopes)
+        .unwrap()
+        .handle_access(access_net)
+        .unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access_fs);
+    assert_eq!(ruleset.actual_handled_fs, access_fs);
+    assert_eq!(ruleset.requested_handled_net, access_net);
+    assert_eq!(ruleset.actual_handled_net, access_net);
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, Scope::EMPTY);
+
+    // Tests composition (binary OR) of handled accesses and scopes.
+    let ruleset = Ruleset::from(ABI::V6)
+        .handle_access(access_fs)
+        .unwrap()
+        .scope(scopes)
+        .unwrap()
+        .handle_access(access_net)
+        .unwrap();
+    assert_eq!(ruleset.requested_handled_fs, access_fs);
+    assert_eq!(ruleset.actual_handled_fs, access_fs);
+    assert_eq!(ruleset.requested_handled_net, access_net);
+    assert_eq!(ruleset.actual_handled_net, access_net);
+    assert_eq!(ruleset.requested_scoped, scopes);
+    assert_eq!(ruleset.actual_scoped, scopes);
+}
+
 impl OptionCompatLevelMut for RulesetCreated {
     fn as_option_compat_level_mut(&mut self) -> &mut Option<CompatLevel> {
         &mut self.compat.level
@@ -479,7 +586,10 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
             match self_ref.compat.state {
                 CompatState::Init | CompatState::No | CompatState::Dummy => Ok(self),
                 CompatState::Full | CompatState::Partial => match unsafe {
-                    uapi::landlock_add_rule(self_ref.fd, T::TYPE_ID, compat_rule.as_ptr(), 0)
+                    #[cfg(test)]
+                    assert!(self_ref.fd.is_some());
+                    let fd = self_ref.fd.as_ref().map(|f| f.as_raw_fd()).unwrap_or(-1);
+                    uapi::landlock_add_rule(fd, T::TYPE_ID, compat_rule.as_ptr(), 0)
                 } {
                     0 => Ok(self),
                     _ => Err(AddRuleError::<U>::AddRuleCall {
@@ -594,7 +704,7 @@ pub trait RulesetCreatedAttr: Sized + AsMut<RulesetCreated> + Compatible {
 /// Ruleset created with [`Ruleset::create()`].
 #[cfg_attr(test, derive(Debug))]
 pub struct RulesetCreated {
-    fd: RawFd,
+    fd: Option<OwnedFd>,
     no_new_privs: bool,
     pub(crate) requested_handled_fs: AccessFs,
     pub(crate) requested_handled_net: AccessNet,
@@ -602,7 +712,7 @@ pub struct RulesetCreated {
 }
 
 impl RulesetCreated {
-    pub(crate) fn new(ruleset: Ruleset, fd: RawFd) -> Self {
+    pub(crate) fn new(ruleset: Ruleset, fd: Option<OwnedFd>) -> Self {
         // The compatibility state is initialized by Ruleset::create().
         #[cfg(test)]
         assert!(!matches!(ruleset.compat.state, CompatState::Init));
@@ -672,7 +782,11 @@ impl RulesetCreated {
                     no_new_privs: enforced_nnp,
                 }),
                 CompatState::Full | CompatState::Partial => {
-                    match unsafe { uapi::landlock_restrict_self(self.fd, 0) } {
+                    #[cfg(test)]
+                    assert!(self.fd.is_some());
+                    // Does not consume ruleset FD, which will be automatically closed after this block.
+                    let fd = self.fd.as_ref().map(|f| f.as_raw_fd()).unwrap_or(-1);
+                    match unsafe { uapi::landlock_restrict_self(fd, 0) } {
                         0 => {
                             self.compat.update(CompatState::Full);
                             Ok(RestrictionStatus {
@@ -697,13 +811,7 @@ impl RulesetCreated {
     /// On error, returns [`std::io::Error`].
     pub fn try_clone(&self) -> std::io::Result<Self> {
         Ok(RulesetCreated {
-            fd: match self.fd {
-                -1 => -1,
-                self_fd => match unsafe { libc::fcntl(self_fd, libc::F_DUPFD_CLOEXEC, 0) } {
-                    dup_fd if dup_fd >= 0 => dup_fd,
-                    _ => return Err(Error::last_os_error()),
-                },
-            },
+            fd: self.fd.as_ref().map(|f| f.try_clone()).transpose()?,
             no_new_privs: self.no_new_privs,
             requested_handled_fs: self.requested_handled_fs,
             requested_handled_net: self.requested_handled_net,
@@ -712,20 +820,21 @@ impl RulesetCreated {
     }
 }
 
-impl Drop for RulesetCreated {
-    fn drop(&mut self) {
-        if self.fd >= 0 {
-            unsafe { close(self.fd) };
-        }
+impl From<RulesetCreated> for Option<OwnedFd> {
+    fn from(ruleset: RulesetCreated) -> Self {
+        ruleset.fd
     }
 }
 
-impl IntoRawFd for RulesetCreated {
-    fn into_raw_fd(self) -> RawFd {
-        let fd = self.fd;
-        std::mem::forget(self);
-        fd
-    }
+#[test]
+fn ruleset_created_ownedfd_none() {
+    let ruleset = Ruleset::from(ABI::Unsupported)
+        .handle_access(AccessFs::Execute)
+        .unwrap()
+        .create()
+        .unwrap();
+    let fd: Option<OwnedFd> = ruleset.into();
+    assert!(fd.is_none());
 }
 
 impl AsMut<RulesetCreated> for RulesetCreated {
@@ -874,6 +983,7 @@ fn ruleset_unsupported() {
         }
     );
 
+    // Missing handled access because of the compatibility level.
     matches!(
         Ruleset::from(ABI::Unsupported)
             // HardRequirement for Ruleset.
@@ -883,6 +993,16 @@ fn ruleset_unsupported() {
         RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
     );
 
+    // Missing scope access because of the compatibility level.
+    matches!(
+        Ruleset::from(ABI::Unsupported)
+            // HardRequirement for Ruleset.
+            .set_compatibility(CompatLevel::HardRequirement)
+            .scope(Scope::Signal)
+            .unwrap_err(),
+        RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
+    );
+
     assert_eq!(
         Ruleset::from(ABI::Unsupported)
             .handle_access(AccessFs::Execute)
@@ -938,6 +1058,7 @@ fn ruleset_unsupported() {
         }
     );
 
+    // Checks empty handled access with moot ruleset.
     assert!(matches!(
         Ruleset::from(ABI::Unsupported)
             // Empty access-rights
@@ -950,12 +1071,13 @@ fn ruleset_unsupported() {
 
     assert!(matches!(
         Ruleset::from(ABI::Unsupported)
-            // No handle_access() call.
+            // No handle_access() or scope call.
             .create()
             .unwrap_err(),
         RulesetError::CreateRuleset(CreateRulesetError::MissingHandledAccess)
     ));
 
+    // Checks empty handled access with minimal ruleset.
     assert!(matches!(
         Ruleset::from(ABI::V1)
             // Empty access-rights
@@ -966,6 +1088,26 @@ fn ruleset_unsupported() {
         )))
     ));
 
+    // Checks empty scope with moot ruleset.
+    assert!(matches!(
+        Ruleset::from(ABI::Unsupported)
+            .scope(Scope::from_all(ABI::Unsupported))
+            .unwrap_err(),
+        RulesetError::HandleAccesses(HandleAccessesError::Scope(HandleAccessError::Compat(
+            CompatError::Access(AccessError::Empty)
+        )))
+    ));
+
+    // Checks empty scope with minimal ruleset.
+    assert!(matches!(
+        Ruleset::from(ABI::V1)
+            .scope(Scope::from_all(ABI::Unsupported))
+            .unwrap_err(),
+        RulesetError::HandleAccesses(HandleAccessesError::Scope(HandleAccessError::Compat(
+            CompatError::Access(AccessError::Empty)
+        )))
+    ));
+
     // Tests inconsistency between the ruleset handled access-rights and the rule access-rights.
     for handled_access in &[
         make_bitflags!(AccessFs::{Execute | WriteFile}),
@@ -976,7 +1118,7 @@ fn ruleset_unsupported() {
             .unwrap();
         // Fakes a call to create() to test without involving the kernel (i.e. no
         // landlock_ruleset_create() call).
-        let ruleset_created = RulesetCreated::new(ruleset, -1);
+        let ruleset_created = RulesetCreated::new(ruleset, None);
         assert!(matches!(
             ruleset_created
                 .add_rule(PathBeneath::new(
