@@ -54,7 +54,7 @@ macro_rules! xpath {
 pub const PATH_MAX: usize = 4096;
 
 /// A safe default size to use for paths.
-pub const PATH_MIN: usize = 128;
+pub const PATH_MIN: usize = 64;
 
 // This pointer is confined by seccomp for use with openat(2) for getdir_long().
 static DOTDOT: Lazy<u64> = Lazy::new(|| c"..".as_ptr() as *const libc::c_char as u64);
@@ -362,9 +362,11 @@ impl XPath {
         //
         // SAFETY: Prevent accessing file names which may be misinterpreted by shells.
         // Note, we skip checking procfs so we don't prevent pipe/socket access
-        // unintentionally.
+        // unintentionally. Similarly we skip checking memory fd names which do not
+        // represent actual file paths.
+        let is_mfd = matches!(file_type, Some(FileType::Mfd));
         let is_proc_dir = self.starts_with(b"/proc");
-        if !is_proc_dir && safe_name && self.check_name().is_err() {
+        if safe_name && !is_mfd && !is_proc_dir && self.check_name().is_err() {
             return Err(Errno::EINVAL);
         }
         // END OF RESTRICTION 2
