@@ -23,7 +23,7 @@ use std::{
 };
 
 use ahash::AHashMap;
-use libc::{c_void, size_t};
+use libc::c_void;
 use libseccomp::ScmpSyscall;
 use nix::{
     errno::Errno,
@@ -32,6 +32,8 @@ use nix::{
 };
 use once_cell::sync::Lazy;
 
+use crate::config::HAVE_MADV_GUARD;
+
 // Note mseal may not be available,
 // and libc::SYS_mseal may not be defined.
 // Therefore we query the number using libseccomp.
@@ -49,14 +51,53 @@ pub fn check_mseal_support() -> bool {
     Errno::result(unsafe { libc::syscall(*SYS_MSEAL, 0, 0, 0) }).is_ok()
 }
 
+/// Check for madvise(2) MADV_GUARD_{INSTALL,REMOVE} support.
+pub fn check_madvise_guard_support() -> bool {
+    *HAVE_MADV_GUARD
+}
+
+/// Safe wrapper for mseal(2) system call.
+pub fn mseal(addr: NonNull<c_void>, len: usize) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe { libc::syscall(*SYS_MSEAL, addr, len, 0) }).map(drop)
+}
+
+/// Safe wrapper for mprotect(2) with PROT_NONE.
+pub fn mprotect_none(addr: NonNull<c_void>, len: usize) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    unsafe { mprotect(addr, len, ProtFlags::PROT_NONE) }
+}
+
+/// Safe wrapper for mprotect(2) with PROT_READ.
+pub fn mprotect_readonly(addr: NonNull<c_void>, len: usize) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    unsafe { mprotect(addr, len, ProtFlags::PROT_READ) }
+}
+
+// Libc does not yet define MADV_GUARD_{INSTALL,REMOVE}.
+const MADV_GUARD_INSTALL: i32 = 102;
+const MADV_GUARD_REMOVE: i32 = 103;
+
+/// Safe wrapper for madvise(2) MADV_GUARD_INSTALL.
+pub fn madvise_guard_install(addr: NonNull<c_void>, len: usize) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe { libc::madvise(addr.as_ptr(), len, MADV_GUARD_INSTALL) }).map(drop)
+}
+
+/// Safe wrapper for madvise(2) MADV_GUARD_REMOVE.
+pub fn madvise_guard_remove(addr: NonNull<c_void>, len: usize) -> Result<(), Errno> {
+    // SAFETY: In libc we trust.
+    Errno::result(unsafe { libc::madvise(addr.as_ptr(), len, MADV_GUARD_REMOVE) }).map(drop)
+}
+
 /// A box for a single `T` in a page-aligned, read/write anonymous mapping,
 /// which can be sealed to read-only. Supports uninitialized allocation,
 /// raw pointers, and zero-boilerplate conversion from `MaybeUninit`.
 #[derive(Debug)]
 pub struct SealBox<T> {
     map_ptr: NonNull<c_void>,
-    map_size: usize,
-    map_zero: bool, // If true, this is a guard page.
+    map_len: usize,
+    map_nul: bool, // If true, this is a guard page.
     _marker: PhantomData<T>,
 }
 
@@ -68,14 +109,14 @@ impl<T> SealBox<T> {
         let page = sysconf(SysconfVar::PAGE_SIZE)?.ok_or(Errno::EINVAL)? as usize;
 
         let size = mem::size_of::<T>();
-        let (map_size, map_prot) = if size == 0 {
+        let (map_len, map_prot) = if size == 0 {
             // SAFETY: Zero-size, map a guard page.
             (page, ProtFlags::PROT_NONE)
         } else {
-            let map_size = size.checked_next_multiple_of(page).ok_or(Errno::EINVAL)?;
-            (map_size, ProtFlags::PROT_READ | ProtFlags::PROT_WRITE)
+            let map_len = size.checked_next_multiple_of(page).ok_or(Errno::EINVAL)?;
+            (map_len, ProtFlags::PROT_READ | ProtFlags::PROT_WRITE)
         };
-        let nn = NonZeroUsize::new(map_size).ok_or(Errno::EINVAL)?;
+        let nn = NonZeroUsize::new(map_len).ok_or(Errno::EINVAL)?;
 
         // SAFETY: valid `length` and flags guaranteed.
         let map_ptr = unsafe { mmap_anonymous(None, nn, map_prot, MapFlags::MAP_PRIVATE)? };
@@ -90,36 +131,33 @@ impl<T> SealBox<T> {
 
         Ok(SealBox {
             map_ptr,
-            map_size,
-            map_zero: size == 0,
+            map_len,
+            map_nul: size == 0,
             _marker: PhantomData,
         })
     }
 
     /// Seal the mapping and return a read-only `Sealed<T>`.
     pub fn seal(self) -> Result<Sealed<T>, Errno> {
-        let addr = self.map_ptr.as_ptr() as *mut u8;
-        let len = self.map_size as size_t;
-
-        if !self.map_zero {
-            // SAFETY: `map_ptr` is page-aligned and `len` is page-aligned.
-            unsafe { mprotect(self.map_ptr, len, ProtFlags::PROT_READ)? };
+        if !self.map_nul {
+            mprotect_readonly(self.map_ptr, self.map_len)?;
+        } else if check_madvise_guard_support() {
+            madvise_guard_install(self.map_ptr, self.map_len)?;
         }
 
-        // SAFETY: raw syscall on valid `addr` and `len`.
         // ENOSYS: mseal(2) is not implemented (Linux>=6.10)
         // EPERM: Sealing is supported only on 64-bit CPUs, 32-bit is not supported.
-        match Errno::result(unsafe { libc::syscall(*SYS_MSEAL, addr, len, 0) }) {
+        match mseal(self.map_ptr, self.map_len) {
             Ok(_) | Err(Errno::EPERM | Errno::ENOSYS) => {}
             Err(errno) => return Err(errno),
         }
 
-        // SAFETY: mapping is now read-only and will live for process lifetime.
+        // SAFETY: Mapping is now read-only and will live for process lifetime.
         let ptr = self.map_ptr.as_ptr().cast::<T>();
         let sealed = Sealed {
             ptr,
             map_ptr: self.map_ptr,
-            map_size: self.map_size,
+            map_len: self.map_len,
         };
 
         mem::forget(self);
@@ -152,22 +190,22 @@ impl<T> SealBox<MaybeUninit<T>> {
         let page = sysconf(SysconfVar::PAGE_SIZE)?.ok_or(Errno::EINVAL)? as usize;
 
         let size = mem::size_of::<T>();
-        let (map_size, map_prot) = if size == 0 {
+        let (map_len, map_prot) = if size == 0 {
             // SAFETY: Zero-size, map a guard page.
             (page, ProtFlags::PROT_NONE)
         } else {
-            let map_size = size.checked_next_multiple_of(page).ok_or(Errno::EINVAL)?;
-            (map_size, ProtFlags::PROT_READ | ProtFlags::PROT_WRITE)
+            let map_len = size.checked_next_multiple_of(page).ok_or(Errno::EINVAL)?;
+            (map_len, ProtFlags::PROT_READ | ProtFlags::PROT_WRITE)
         };
-        let nn = NonZeroUsize::new(map_size).ok_or(Errno::EINVAL)?;
+        let nn = NonZeroUsize::new(map_len).ok_or(Errno::EINVAL)?;
 
         // SAFETY: valid `length` and flags guaranteed.
         let map_ptr = unsafe { mmap_anonymous(None, nn, map_prot, MapFlags::MAP_PRIVATE)? };
 
         Ok(SealBox {
             map_ptr,
-            map_size,
-            map_zero: size == 0,
+            map_len,
+            map_nul: size == 0,
             _marker: PhantomData,
         })
     }
@@ -185,13 +223,13 @@ impl<T> SealBox<MaybeUninit<T>> {
     /// You must have fully initialized the `T` in the mapping.
     pub unsafe fn assume_init(self) -> SealBox<T> {
         let map_ptr = self.map_ptr;
-        let map_size = self.map_size;
-        let map_zero = self.map_zero;
+        let map_len = self.map_len;
+        let map_nul = self.map_nul;
         mem::forget(self);
         SealBox {
             map_ptr,
-            map_size,
-            map_zero,
+            map_len,
+            map_nul,
             _marker: PhantomData,
         }
     }
@@ -208,7 +246,7 @@ impl<T> Deref for SealBox<T> {
 impl<T> Drop for SealBox<T> {
     fn drop(&mut self) {
         // SAFETY: valid mapping to unmap
-        let _ = unsafe { munmap(self.map_ptr, self.map_size) };
+        let _ = unsafe { munmap(self.map_ptr, self.map_len) };
     }
 }
 
@@ -216,7 +254,7 @@ impl<T> Drop for SealBox<T> {
 pub struct Sealed<T> {
     ptr: *const T,
     map_ptr: NonNull<c_void>,
-    map_size: usize,
+    map_len: usize,
 }
 
 impl<T> Sealed<T> {
@@ -227,7 +265,7 @@ impl<T> Sealed<T> {
 
     /// Get the underlying mapping pointer and size.
     pub fn mapping(&self) -> (NonNull<c_void>, usize) {
-        (self.map_ptr, self.map_size)
+        (self.map_ptr, self.map_len)
     }
 }
 
@@ -244,7 +282,7 @@ impl<T> Deref for Sealed<T> {
 #[derive(Debug)]
 pub struct SealBoxSlice<E> {
     map_ptr: NonNull<c_void>,
-    map_size: usize,
+    map_len: usize,
     len: usize,
     _marker: PhantomData<E>,
 }
@@ -271,7 +309,7 @@ impl<E> SealBoxSlice<E> {
                 unsafe { mmap_anonymous(None, nn, ProtFlags::PROT_NONE, MapFlags::MAP_PRIVATE) }?;
             return Ok(SealBoxSlice {
                 map_ptr,
-                map_size: page,
+                map_len: page,
                 len: 0,
                 _marker: PhantomData,
             });
@@ -283,10 +321,10 @@ impl<E> SealBoxSlice<E> {
         }
         let data_size = elem_size.checked_mul(len).ok_or(Errno::EINVAL)?;
 
-        let map_size = data_size
+        let map_len = data_size
             .checked_next_multiple_of(page)
             .ok_or(Errno::EINVAL)?;
-        let nn = NonZeroUsize::new(map_size).ok_or(Errno::EINVAL)?;
+        let nn = NonZeroUsize::new(map_len).ok_or(Errno::EINVAL)?;
 
         // SAFETY: valid `length` and flags guaranteed.
         let map_ptr = unsafe {
@@ -306,7 +344,7 @@ impl<E> SealBoxSlice<E> {
 
         Ok(SealBoxSlice {
             map_ptr,
-            map_size,
+            map_len,
             len,
             _marker: PhantomData,
         })
@@ -324,32 +362,29 @@ impl<E> SealBoxSlice<E> {
 
     /// Seal the mapping and return a read-only `SealedSlice<E>`.
     pub fn seal(self) -> Result<SealedSlice<E>, Errno> {
-        let addr = self.map_ptr.as_ptr() as *mut u8;
-        let len_pages = self.map_size as size_t;
-
         if self.len > 0 {
             // Non-zero sized slice, map read-only.
-            // SAFETY: page-aligned.
-            unsafe { mprotect(self.map_ptr, len_pages, ProtFlags::PROT_READ)? };
+            mprotect_readonly(self.map_ptr, self.map_len)?;
         }
 
-        // SAFETY: raw syscall on valid `addr` and `len_pages`.
         // ENOSYS: mseal(2) is not implemented (Linux>=6.10)
         // EPERM: Sealing is supported only on 64-bit CPUs, 32-bit is not supported.
-        match Errno::result(unsafe { libc::syscall(*SYS_MSEAL, addr, len_pages, 0) }) {
+        match mseal(self.map_ptr, self.map_len) {
             Ok(_) | Err(Errno::EPERM | Errno::ENOSYS) => {}
             Err(errno) => return Err(errno),
         }
 
-        // SAFETY: mapping is now read-only and will live for process lifetime.
+        // SAFETY: Mapping is now read-only and will live for process lifetime.
         let ptr = self.map_ptr.as_ptr().cast::<E>();
         let sealed = SealedSlice {
             ptr,
             len: self.len,
             map_ptr: self.map_ptr,
-            map_size: self.map_size,
+            map_len: self.map_len,
         };
+
         mem::forget(self);
+
         Ok(sealed)
     }
 }
@@ -357,7 +392,7 @@ impl<E> SealBoxSlice<E> {
 impl<E> Drop for SealBoxSlice<E> {
     fn drop(&mut self) {
         // SAFETY: valid mapping to unmap.
-        let _ = unsafe { munmap(self.map_ptr, self.map_size) };
+        let _ = unsafe { munmap(self.map_ptr, self.map_len) };
     }
 }
 
@@ -366,7 +401,7 @@ pub struct SealedSlice<E> {
     ptr: *const E,
     len: usize,
     map_ptr: NonNull<c_void>,
-    map_size: usize,
+    map_len: usize,
 }
 
 impl<E> SealedSlice<E> {
@@ -387,7 +422,7 @@ impl<E> SealedSlice<E> {
 
     /// Get the underlying mapping pointer and size.
     pub fn mapping(&self) -> (NonNull<c_void>, usize) {
-        (self.map_ptr, self.map_size)
+        (self.map_ptr, self.map_len)
     }
 }
 
@@ -476,7 +511,7 @@ impl<K: Hash + Eq + Ord + Clone, V: Clone> SealableSlice<(K, V)> for AHashMap<K,
     }
 }
 
-/// A `Copy` value which can be either unsealed or sealed in a read‐only mapping.
+/// A `Copy` value which can be either unsealed or sealed in a read-only mapping.
 pub enum Sealable<T: Copy> {
     /// The unsealed state.
     Unsealed(T),
@@ -490,7 +525,7 @@ impl<T: Copy> Sealable<T> {
         Sealable::Unsealed(val)
     }
 
-    /// Seal the value into a read‐only mmap.
+    /// Seal the value into a read-only mmap.
     pub fn seal(self) -> Result<Self, Errno> {
         match self {
             Sealable::Unsealed(v) => {
@@ -1134,11 +1169,11 @@ impl<'a, T> IntoIterator for &'a SealableVecDeque<T> {
     }
 }
 
-/// A HashMap‐like container that can be sealed into a read‐only mmap of sorted entries.
+/// A HashMap-like container that can be sealed into a read-only mmap of sorted entries.
 pub enum SealableHashMap<K, V> {
     /// The unsealed, growable `HashMap<K, V>` state.
     Unsealed(HashMap<K, V>),
-    /// The sealed, read‐only slice state of sorted `(K, V)` pairs.
+    /// The sealed, read-only slice state of sorted `(K, V)` pairs.
     Sealed(SealedSlice<(K, V)>),
 }
 
@@ -1153,7 +1188,7 @@ impl<K, V> SealableHashMap<K, V> {
         SealableHashMap::Unsealed(HashMap::with_capacity(cap))
     }
 
-    /// Seal into a read‐only mmap of sorted `(K, V)` pairs.
+    /// Seal into a read-only mmap of sorted `(K, V)` pairs.
     ///
     /// In the sealed state, lookups use binary search on the sorted slice.
     pub fn seal(self) -> Result<Self, Errno>
@@ -1209,7 +1244,7 @@ impl<K, V> SealableHashMap<K, V> {
         self.get(key).is_some()
     }
 
-    /// Inserts a key‐value pair into the map.
+    /// Inserts a key-value pair into the map.
     ///
     /// Returns the old value if the key was already present.
     /// In the sealed state, returns `Err(EPERM)`.
@@ -1353,11 +1388,11 @@ where
     }
 }
 
-/// A AHashMap‐like container that can be sealed into a read‐only mmap of sorted entries.
+/// A AHashMap-like container that can be sealed into a read-only mmap of sorted entries.
 pub enum SealableAHashMap<K, V> {
     /// The unsealed, growable `AHashMap<K, V>` state.
     Unsealed(AHashMap<K, V>),
-    /// The sealed, read‐only slice state of sorted `(K, V)` pairs.
+    /// The sealed, read-only slice state of sorted `(K, V)` pairs.
     Sealed(SealedSlice<(K, V)>),
 }
 
@@ -1372,7 +1407,7 @@ impl<K, V> SealableAHashMap<K, V> {
         SealableAHashMap::Unsealed(AHashMap::with_capacity(cap))
     }
 
-    /// Seal into a read‐only mmap of sorted `(K, V)` pairs.
+    /// Seal into a read-only mmap of sorted `(K, V)` pairs.
     ///
     /// In the sealed state, lookups use binary search on the sorted slice.
     pub fn seal(self) -> Result<Self, Errno>
@@ -1428,7 +1463,7 @@ impl<K, V> SealableAHashMap<K, V> {
         self.get(key).is_some()
     }
 
-    /// Inserts a key‐value pair into the map.
+    /// Inserts a key-value pair into the map.
     ///
     /// Returns the old value if the key was already present.
     /// In the sealed state, returns `Err(EPERM)`.
@@ -1621,7 +1656,12 @@ unsafe impl<K: Send, V: Sync> Sync for SealableAHashMap<K, V> {}
 mod tests {
     use nix::{
         errno::Errno,
-        sys::mman::{madvise, mremap, MRemapFlags, MmapAdvise, ProtFlags},
+        sys::{
+            mman::{madvise, mremap, MRemapFlags, MmapAdvise, ProtFlags},
+            signal::Signal,
+            wait::{waitpid, WaitStatus},
+        },
+        unistd::{fork, ForkResult},
     };
 
     use super::*;
@@ -1874,9 +1914,9 @@ mod tests {
         for s in sealed.iter() {
             assert_eq!(s, &"x".repeat(50));
         }
-        let (_ptr, map_size) = sealed.mapping();
-        assert_eq!(map_size % page, 0);
-        assert!(map_size >= len * mem::size_of::<String>());
+        let (_ptr, map_len) = sealed.mapping();
+        assert_eq!(map_len % page, 0);
+        assert!(map_len >= len * mem::size_of::<String>());
     }
 
     #[test]
@@ -2285,4 +2325,291 @@ mod tests {
         let err = sealed.retain(|_, _| true).unwrap_err();
         assert_eq!(err, Errno::EPERM);
     }
+
+    //
+    // MADV_GUARD_{INSTALL,REMOVE} tests.
+    //
+
+    // Helper get system page size.
+    fn page_size() -> usize {
+        sysconf(SysconfVar::PAGE_SIZE)
+            .expect("sysconf PAGE_SIZE")
+            .expect("PAGE_SIZE not defined") as usize
+    }
+
+    // Helper: Map `PAGE_SIZE` bytes of anonymous read-write memory.
+    fn map_region(len: usize) -> NonNull<c_void> {
+        let result = unsafe {
+            mmap_anonymous(
+                None,
+                NonZeroUsize::new(len).unwrap(),
+                ProtFlags::PROT_READ | ProtFlags::PROT_WRITE,
+                MapFlags::MAP_PRIVATE,
+            )
+        };
+
+        assert!(result.is_ok(), "mmap_anonymous:{result:?}");
+        result.unwrap()
+    }
+
+    // Helper: Unmap a region.
+    fn unmap_region(ptr: NonNull<c_void>, len: usize) {
+        let result = unsafe { munmap(ptr, len) };
+        assert!(result.is_ok(), "munmap:{result:?}");
+    }
+
+    #[test]
+    fn test_madv_guard_zero_length_install() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let dummy = NonNull::new(1 as *mut c_void).unwrap();
+        assert_eq!(madvise_guard_install(dummy, 0).unwrap_err(), Errno::EINVAL);
+    }
+
+    #[test]
+    fn test_madv_guard_zero_length_remove() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let dummy = NonNull::new(1 as *mut c_void).unwrap();
+        assert_eq!(madvise_guard_remove(dummy, 0).unwrap_err(), Errno::EINVAL);
+    }
+
+    #[test]
+    fn test_madv_guard_install_idempotent() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page);
+
+        // Attempt first install; skip if guard not supported.
+        match madvise_guard_install(region, page) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(region, page);
+                return;
+            }
+            Err(errno) => panic!("unexpected error on install: {errno}"),
+        }
+
+        // Second install should still succeed.
+        madvise_guard_install(region, page).expect("second install failed");
+
+        // Clean up.
+        let _ = madvise_guard_remove(region, page);
+        unmap_region(region, page);
+    }
+
+    #[test]
+    fn test_madv_guard_remove_idempotent() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page);
+
+        // Must install first; skip if unsupported.
+        match madvise_guard_install(region, page) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(region, page);
+                return;
+            }
+            Err(errno) => panic!("unexpected error on install: {errno}"),
+        }
+
+        // First remove.
+        madvise_guard_remove(region, page).expect("first remove failed");
+
+        // Second remove is a no-op.
+        madvise_guard_remove(region, page).expect("second remove failed");
+
+        unmap_region(region, page);
+    }
+
+    #[test]
+    fn test_madv_guard_remove_without_install() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page);
+
+        match madvise_guard_remove(region, page) {
+            Ok(()) => {}             // treated as no-op.
+            Err(Errno::EINVAL) => {} // OK
+            Err(errno) => panic!("unexpected error: {errno}"),
+        }
+
+        unmap_region(region, page);
+    }
+
+    #[test]
+    fn test_madv_guard_install_partial_fault() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page * 2);
+
+        // Install guard only on first page.
+        match madvise_guard_install(region, page) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(region, page * 2);
+                return;
+            }
+            Err(errno) => panic!("unexpected error: {errno}"),
+        }
+
+        match unsafe { fork() } {
+            Ok(ForkResult::Parent { child }) => match waitpid(child, None).unwrap() {
+                WaitStatus::Signaled(_, Signal::SIGSEGV, _) => {}
+                other => panic!("expected SIGSEGV, got {other:?}"),
+            },
+            Ok(ForkResult::Child) => {
+                unsafe { (region.as_ptr() as *mut u8).write_volatile(0) };
+                std::process::exit(1);
+            }
+            Err(errno) => panic!("fork failed: {errno}"),
+        }
+
+        let _ = madvise_guard_remove(region, page);
+        unmap_region(region, page * 2);
+    }
+
+    #[test]
+    fn test_madv_guard_install_partial_no_fault() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page * 2);
+
+        match madvise_guard_install(region, page) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(region, page * 2);
+                return;
+            }
+            Err(errno) => panic!("unexpected error: {errno}"),
+        }
+
+        // Writing to second page should succeed.
+        match unsafe { fork() } {
+            Ok(ForkResult::Parent { child }) => match waitpid(child, None).unwrap() {
+                WaitStatus::Exited(_, 0) => {}
+                other => panic!("expected exit(0), got {other:?}"),
+            },
+            Ok(ForkResult::Child) => {
+                unsafe { (region.as_ptr().add(page) as *mut u8).write_volatile(42) };
+                std::process::exit(0);
+            }
+            Err(errno) => panic!("fork failed: {errno}"),
+        }
+
+        let _ = madvise_guard_remove(region, page);
+        unmap_region(region, page * 2);
+    }
+
+    #[test]
+    fn test_madv_guard_remove_partial_no_fault() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let region = map_region(page * 2);
+
+        // Install guard on both pages, then remove on second page.
+        match madvise_guard_install(region, page * 2) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(region, page * 2);
+                return;
+            }
+            Err(errno) => panic!("unexpected error: {errno}"),
+        }
+        let _ = madvise_guard_remove(
+            unsafe { NonNull::new_unchecked(region.as_ptr().add(page)) },
+            page,
+        );
+
+        // Writing to second page must succeed.
+        match unsafe { fork() } {
+            Ok(ForkResult::Parent { child }) => match waitpid(child, None).unwrap() {
+                WaitStatus::Exited(_, 0) => {}
+                other => panic!("expected exit(0), got {other:?}"),
+            },
+            Ok(ForkResult::Child) => {
+                unsafe { (region.as_ptr().add(page) as *mut u8).write_volatile(99) };
+                std::process::exit(0);
+            }
+            Err(errno) => panic!("fork failed: {errno}"),
+        }
+
+        let _ = madvise_guard_remove(region, page * 2);
+        unmap_region(region, page * 2);
+    }
+
+    #[test]
+    fn test_madv_guard_multiple_regions() {
+        if !check_madvise_guard_support() {
+            return;
+        }
+
+        let page = page_size();
+        let r1 = map_region(page);
+        let r2 = map_region(page);
+
+        // Install guard on r1 only.
+        match madvise_guard_install(r1, page) {
+            Ok(()) => {}
+            Err(Errno::EINVAL) => {
+                unmap_region(r1, page);
+                unmap_region(r2, page);
+                return;
+            }
+            Err(errno) => panic!("unexpected error: {errno}"),
+        }
+
+        // r2 should remain writable.
+        match unsafe { fork() } {
+            Ok(ForkResult::Parent { child }) => match waitpid(child, None).unwrap() {
+                WaitStatus::Exited(_, 0) => {}
+                other => panic!("expected exit(0), got {other:?}"),
+            },
+            Ok(ForkResult::Child) => {
+                unsafe { (r2.as_ptr() as *mut u8).write_volatile(7) };
+                std::process::exit(0);
+            }
+            Err(errno) => panic!("fork failed: {errno}"),
+        }
+
+        // But r1 should fault.
+        match unsafe { fork() } {
+            Ok(ForkResult::Parent { child }) => match waitpid(child, None).unwrap() {
+                WaitStatus::Signaled(_, Signal::SIGSEGV, _) => {}
+                other => panic!("expected SIGSEGV, got {other:?}"),
+            },
+            Ok(ForkResult::Child) => {
+                unsafe { (r1.as_ptr() as *mut u8).write_volatile(5) };
+                std::process::exit(1);
+            }
+            Err(errno) => panic!("fork failed: {errno}"),
+        }
+
+        let _ = madvise_guard_remove(r1, page);
+        unmap_region(r1, page);
+        unmap_region(r2, page);
+    }
 }
