@@ -84,7 +84,7 @@ use nix::{
 };
 use sendfd::SendWithFd;
 use syd::{
-    bring_up_loopback, caps,
+    caps,
     compat::{waitid, WaitStatus},
     config::*,
     err::{err2no, SydResult},
@@ -95,7 +95,7 @@ use syd::{
     info,
     landlock_policy::LandlockPolicy,
     log::log_init,
-    nsflags_name,
+    loopback_set_up, nsflags_name,
     path::XPathBuf,
     sandbox::Sandbox,
     syslog::LogLevel,
@@ -267,6 +267,12 @@ fn main() -> SydResult<ExitCode> {
                 println!("{ESYD_SH}");
                 return Ok(ExitCode::SUCCESS);
             }
+            Long("api") => {
+                #[allow(clippy::disallowed_methods)]
+                let api = serde_json::to_string_pretty(&*syd::api::API_SPEC).expect("JSON");
+                println!("{api}");
+                return Ok(ExitCode::SUCCESS);
+            }
             Short('q') => {} // Ignore, must be first!
 
             /*
@@ -431,7 +437,10 @@ fn main() -> SydResult<ExitCode> {
     }
 
     let pty_debug = env::var_os("SYD_PTY_DEBUG").is_some();
-    let pty_child = if sandbox.has_pty() && isatty(std::io::stdin()).unwrap_or(false) {
+    let pty_child = if sandbox.has_pty()
+        && isatty(std::io::stdin()).unwrap_or(false)
+        && isatty(std::io::stdout()).unwrap_or(false)
+    {
         // Step 1: Create a PIDFd of this process and clear O_CLOEXEC.
         let pidfd = pidfd_open(getpid(), libc::PIDFD_NONBLOCK)?;
         set_cloexec(&pidfd, false)?;
@@ -569,17 +578,17 @@ fn main() -> SydResult<ExitCode> {
         cmd.arg(format!("-i{}", stream_parent.as_raw_fd()));
         // proxy/ext/unix has precedence over proxy/ext/host.
         if let Some(ref proxy_ext_unix) = sandbox.proxy_ext_unix {
-            cmd.arg(format!("-o{proxy_ext_unix}"));
+            cmd.arg("-u");
+            cmd.arg(proxy_ext_unix);
         } else {
+            cmd.arg("-o");
             cmd.arg(format!(
-                "-o{}:{}",
+                "{}:{}",
                 sandbox.proxy_ext_addr, sandbox.proxy_ext_port
             ));
         }
-        //cmd.arg(format!("-b{}", PIPE_BUF));
         cmd.env_clear();
         cmd.envs(safe_env);
-        cmd.current_dir("/");
         cmd.process_group(0);
         cmd.stdin(Stdio::null());
         cmd.stdout(Stdio::null());
@@ -588,14 +597,18 @@ fn main() -> SydResult<ExitCode> {
         }
         cmd.spawn()?;
         drop(pidfd);
+
+        let proxy_repr = sandbox
+            .proxy_ext_unix
+            .as_ref()
+            .map(|proxy_ext_unix| proxy_ext_unix.to_string())
+            .unwrap_or_else(|| format!("{}!{}", sandbox.proxy_ext_addr, sandbox.proxy_ext_port));
         if proxy_debug {
             warn!("ctx": "spawn_proxy",
-                "msg": format!("proxy is now forwarding external traffic to {}!{}",
-                    sandbox.proxy_ext_addr, sandbox.proxy_ext_port));
+                "msg": format!("proxy is now forwarding external traffic to {proxy_repr}"));
         } else {
             info!("ctx": "spawn_proxy",
-                "msg": format!("proxy is now forwarding external traffic to {}!{}",
-                    sandbox.proxy_ext_addr, sandbox.proxy_ext_port));
+                "msg": format!("proxy is now forwarding external traffic to {proxy_repr}"));
         }
 
         // Step 4: Pass the other end of the socket-pair
@@ -716,17 +729,18 @@ fn main() -> SydResult<ExitCode> {
     if sandbox.unshare_net() {
         // Set up the loopback interface.
         // Warn on errors and continue.
-        match bring_up_loopback() {
-            Ok(_) => {
-                info!("ctx": "bring_up_loopback",
-                    "msg": "loopback network device is now up");
+        match loopback_set_up() {
+            Ok(loindex) => {
+                info!("ctx": "loopback_set_up",
+                    "msg": format!("loopback network device is up with index:{loindex:#x}"),
+                    "idx": loindex);
             }
             Err(errno) => {
-                error!("ctx": "bring_up_loopback_device",
-                    "msg": format!("bring up loopback network device error: {errno}"),
+                error!("ctx": "loopback_set_up",
+                    "msg": format!("set up loopback network device error: {errno}"),
                     "err": errno as i32);
             }
-        }
+        };
 
         // Handle proxy sandboxing.
         // Warn on errors and continue.
