@@ -44,7 +44,10 @@ mod stdio;
 mod wait;
 mod zombies;
 
-use std::{ffi::CString, os::unix::io::RawFd};
+use std::{
+    ffi::CString,
+    os::{fd::OwnedFd, unix::io::RawFd},
+};
 
 use libseccomp::ScmpFilterContext;
 pub use nix::sys::signal::Signal;
@@ -66,6 +69,7 @@ pub struct Command {
     config: config::Config,
     before_unfreeze: Option<Box<dyn FnMut(u32) -> Result<(), Errno>>>,
     pre_exec: Option<Box<dyn Fn() -> Result<(), Errno>>>,
+    pty_fd: Option<OwnedFd>,
     seccomp_filter: Option<ScmpFilterContext>,
     seccomp_pipefd: ((RawFd, RawFd), (RawFd, RawFd)),
 }
