@@ -23,7 +23,7 @@ use std::{
 };
 
 use bitflags::bitflags;
-use hex::DisplayHex;
+use data_encoding::HEXLOWER;
 use libseccomp::{ScmpAction, ScmpFilterContext, ScmpSyscall};
 use memchr::memchr3;
 use nix::{
@@ -534,7 +534,7 @@ impl WordExp<'_> {
         // SAFETY: hex-encode if expansion is invalid UTF-8.
         let ret = match std::str::from_utf8(&ret) {
             Ok(ret) => ret.to_string(),
-            Err(_) => return Ok(ret.to_lower_hex_string().into()),
+            Err(_) => return Ok(HEXLOWER.encode(&ret).into()),
         };
 
         // SAFETY: do not allow empty replacement.
@@ -544,7 +544,7 @@ impl WordExp<'_> {
 
         // SAFETY: hex-encode if string has non-printables.
         if contains_ascii_unprintable(ret.as_bytes()) {
-            Ok(ret.as_bytes().to_lower_hex_string().into())
+            Ok(HEXLOWER.encode(ret.as_bytes()).into())
         } else {
             Ok(ret.into())
         }
