@@ -9,7 +9,7 @@
 use std::{
     collections::HashMap,
     env,
-    os::fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
+    os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
     sync::{Arc, RwLock},
     thread,
     thread::JoinHandle,
@@ -27,13 +27,13 @@ use crate::{
     config::*,
     err::{err2no, SydResult},
     error,
-    fs::{lock_fd, retry_on_eintr, seal_memfd, FileInfo},
+    fs::{lock_fd, retry_on_eintr, retry_on_intr, seal_memfd, FileInfo},
     hash::{
         aes_ctr_enc, aes_ctr_init, hmac_sha256_feed, hmac_sha256_fini, hmac_sha256_init,
         SydRandomState, BLOCK_SIZE, HMAC_TAG_SIZE, IV, IV_SIZE,
     },
     info,
-    path::XPathBuf,
+    path::{XPath, XPathBuf},
     sandbox::Flags,
     ExportMode,
 };
@@ -62,8 +62,8 @@ impl From<OFlag> for AesMod {
 #[derive(Debug)]
 pub(crate) struct AesVal {
     pub(crate) crypt_fd: RawFd,
-    pub(crate) enc_fd: OwnedFd,
-    pub(crate) iv: IV,
+    pub(crate) enc_fd: Option<OwnedFd>,
+    pub(crate) iv: Option<IV>,
     pub(crate) info: FileInfo,
     pub(crate) mode: AesMod,
     pub(crate) spawned: bool,
@@ -145,9 +145,7 @@ impl AesWorker {
             let finished = my_files.1;
             for (path, v) in &my_files.0 {
                 if !v.spawned {
-                    // SAFETY: crypt_map keys are valid FDs.
-                    let crypt_fd = unsafe { OwnedFd::from_raw_fd(v.crypt_fd) };
-                    paths.push((crypt_fd, path.clone()));
+                    paths.push((v.crypt_fd, path.clone()));
                 }
             }
             drop(my_files);
@@ -174,9 +172,12 @@ impl AesWorker {
 
             for (crypt_fd, crypt_path) in paths.drain(..) {
                 // Spawn a thread to handle the write.
-                let files = Arc::clone(&files);
                 threads.push(Self::spawn(
-                    setup_fds, crypt_fd, crypt_path, files, is_memfd,
+                    setup_fds,
+                    crypt_fd,
+                    &crypt_path,
+                    &files,
+                    is_memfd,
                 ));
             }
 
@@ -193,49 +194,69 @@ impl AesWorker {
 
     fn spawn(
         setup_fds: (RawFd, RawFd),
-        crypt_fd: OwnedFd,
-        crypt_path: XPathBuf,
-        files: AesMap,
+        crypt_fd: RawFd,
+        crypt_path: &XPath,
+        files: &AesMap,
         memfd: bool,
     ) -> JoinHandle<Result<(), Errno>> {
         #[allow(clippy::disallowed_methods)]
-        thread::Builder::new()
-            .name("syd_aes".into())
-            .stack_size(AES_STACK_SIZE)
-            .spawn(move || {
-                // Wait until we take a write lock on the encrypted fd.
-                // This will succeed once all fds owned by the sandbox
-                // process are closed.
-                lock_fd(&crypt_fd, true, true)?;
-                // All good, sync contents to disk.
-                Self::sync(setup_fds, crypt_fd, crypt_path, files, memfd)
-            })
-            .unwrap()
+        retry_on_intr(|| {
+            let crypt_path = XPathBuf::from(crypt_path);
+            let files = Arc::clone(files);
+
+            thread::Builder::new()
+                .name("syd_aes".into())
+                .stack_size(AES_STACK_SIZE)
+                .spawn(move || {
+                    // SAFETY: crypt_map keys are valid FDs.
+                    let crypt_fd = unsafe { OwnedFd::from_raw_fd(crypt_fd) };
+
+                    // Wait until we take a write lock on the encrypted fd.
+                    // This will succeed once all fds owned by the sandbox
+                    // process are closed.
+                    lock_fd(&crypt_fd, true, true)?;
+
+                    // All good, sync contents to disk.
+                    let result = Self::sync(setup_fds, &crypt_fd, &crypt_path, &files, memfd);
+
+                    // Safe to remove file entry now.
+                    let mut files = files.write().unwrap_or_else(|err| err.into_inner());
+                    #[allow(clippy::disallowed_methods)]
+                    files.0.remove(&crypt_path).map(drop).unwrap();
+
+                    // Close the encrypted FD.
+                    drop(crypt_fd);
+
+                    result
+                })
+                .map_err(|err| err2no(&err))
+        })
+        .expect("spawn syd_aes")
     }
 
     #[allow(clippy::arithmetic_side_effects)]
     #[allow(clippy::cognitive_complexity)]
     #[allow(clippy::disallowed_methods)]
-    fn sync(
+    fn sync<Fd: AsFd>(
         setup_fds: (RawFd, RawFd),
-        crypt_fd: OwnedFd,
-        crypt_path: XPathBuf,
-        files: AesMap,
+        crypt_fd: Fd,
+        crypt_path: &XPath,
+        files: &AesMap,
         memfd: bool,
     ) -> Result<(), Errno> {
-        let (aes_fd, mac_fd) = setup_fds;
-
-        let (enc_fd, file_mode, mut iv) = {
-            let mut files = files.write().unwrap_or_else(|err| err.into_inner());
-            let v = files.0.remove(&crypt_path).unwrap();
-            (v.enc_fd, v.mode, v.iv)
-        };
-
         // Seal memfd to ensure no further writes happen.
         if memfd {
             seal_memfd(&crypt_fd).unwrap();
         }
 
+        let (aes_fd, mac_fd) = setup_fds;
+        let (enc_fd, file_mode, mut iv) = {
+            let mut files = files.write().unwrap_or_else(|err| err.into_inner());
+            let v = files.0.get_mut(crypt_path).unwrap();
+            #[allow(clippy::disallowed_methods)]
+            (v.enc_fd.take().unwrap(), v.mode, v.iv.take().unwrap())
+        };
+
         // Nothing to do if file was readonly.
         let mut is_append = match file_mode {
             AesMod::Read => return Ok(()),
@@ -721,6 +742,22 @@ impl AesWorker {
             }
         }
 
+        // Deny installing new signal handlers for {rt_,}sigaction(2).
+        for sysname in ["sigaction", "rt_sigaction"] {
+            let syscall = match ScmpSyscall::from_name(sysname) {
+                Ok(syscall) => syscall,
+                Err(_) => {
+                    crate::info!("ctx": "confine", "op": "allow_main_syscall",
+                        "msg": format!("invalid or unsupported syscall {sysname}"));
+                    continue;
+                }
+            };
+
+            // Installs a signal handler if first argument is non-NULL.
+            // We deny this case, but allow returning the current handler.
+            ctx.add_rule_conditional(ScmpAction::Allow, syscall, &[scmp_cmp!($arg1 == 0)])?;
+        }
+
         // Allow safe system calls.
         for sysname in AES_SYSCALLS {
             match ScmpSyscall::from_name(sysname) {
