@@ -10,7 +10,7 @@ use std::{
     borrow::Cow,
     collections::HashMap,
     hash::{Hash, Hasher},
-    io::{Read, Write},
+    io::{IoSlice, IoSliceMut},
     os::{
         fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
         unix::net::{UnixListener, UnixStream},
@@ -23,16 +23,20 @@ use std::{
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use memchr::{arch::all::is_equal, memchr};
 use nix::{
+    cmsg_space,
     errno::Errno,
     fcntl::OFlag,
     poll::PollTimeout,
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
         socket::{
-            accept4, bind, listen, socket, AddressFamily, Backlog, SockFlag, SockType, UnixAddr,
+            accept4, bind, listen, recvmsg, sendmsg, setsockopt, socket, sockopt::PassCred,
+            AddressFamily, Backlog, ControlMessageOwned, MsgFlags, SockFlag, SockType, UnixAddr,
+            UnixCredentials,
         },
         stat::{umask, Mode},
     },
+    unistd::{Gid, Uid},
 };
 
 use crate::{
@@ -48,6 +52,8 @@ use crate::{
 };
 
 const IPC_ACK: &[u8] = b"{\"err\":0,\"msg\":\"ACK\"}\n";
+const IPC_AUTH: &[u8] = b"{\"err\":13,\"msg\":\"AUTH\"}\n";
+const IPC_RATE: &[u8] = b"{\"err\":7,\"msg\":\"RATE\"}\n";
 const IPC_PINK: &[u8] = b"{\"err\":0,\
 \"msg\":\"Change return success. Going and coming without error. Action brings good fortune.\"}\n";
 const IPC_PONG: &[u8] = b"{\"err\":0,\"msg\":\"PONG\"}\n";
@@ -66,6 +72,8 @@ pub(crate) struct IpcWorker {
     sandbox: Option<Arc<RwLock<Sandbox>>>,
     // Sandbox flags specified at startup.
     flags: Flags,
+    // Credentials for authentication.
+    creds: (Option<Uid>, Option<Gid>),
 }
 
 // Prompt modes supported (similar to HAProxy)
@@ -79,7 +87,8 @@ enum PromptMode {
     Prompt,
 }
 
-/// Per-connection state: buffering incoming bytes, pending outgoing bytes, and tracking mode.
+/// Per-connection state: buffering incoming bytes,
+/// pending outgoing bytes, and tracking mode.
 struct Connection {
     /// Underlying UNIX-stream socket.
     stream: UnixStream,
@@ -125,6 +134,19 @@ impl Connection {
         }
     }
 
+    // Change epoll(7) interests.
+    fn ctl(&self, epoll: &Epoll, flags: EpollFlags) -> Result<(), Errno> {
+        let fd = self.stream.as_raw_fd();
+
+        #[allow(clippy::cast_sign_loss)]
+        let event = libc::epoll_event {
+            events: flags.bits() as u32,
+            u64: fd as u64,
+        };
+
+        epoll_ctl_mod_safe(&epoll.0, fd, event)
+    }
+
     // Append data to the write buffer.
     #[inline]
     fn enqueue_response(&mut self, data: &[u8]) {
@@ -145,9 +167,10 @@ impl IpcWorker {
     //
     // This does not bind or listen yet; only stores the path.
     // Returns immediately without error.
-    pub(crate) fn new(addr: &XPath, flags: Flags) -> Self {
+    pub(crate) fn new(addr: &XPath, uid: Option<Uid>, gid: Option<Gid>, flags: Flags) -> Self {
         IpcWorker {
             flags,
+            creds: (uid, gid),
             addr: Some(addr.to_owned()),
             epoll: None,
             sock: None,
@@ -179,7 +202,9 @@ impl IpcWorker {
         // SAFETY: Randomize to make fd reuse harder.
         let sock_fd = duprand(sock.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
-            std::env::set_var(ENV_IPC_UNIX_FD, fd.to_string());
+            let mut fd_str = itoa::Buffer::new();
+            let fd_str = fd_str.format(fd);
+            std::env::set_var(ENV_IPC_UNIX_FD, fd_str);
 
             // SAFETY: duprand returns a valid FD on success.
             unsafe { OwnedFd::from_raw_fd(fd) }
@@ -187,6 +212,9 @@ impl IpcWorker {
         drop(sock);
         let sock = sock_fd;
 
+        // Set SO_PASSCRED for authentication.
+        setsockopt(&sock, PassCred, &true)?;
+
         // SAFETY: Ensure socket is created with sane permissions.
         let umask_orig = umask(Mode::from_bits_truncate(0o077));
         let result = bind(sock.as_raw_fd(), &addr);
@@ -201,7 +229,9 @@ impl IpcWorker {
         // SAFETY: Randomize the epoll fd to make fd reuse harder.
         let epoll_fd = duprand(epoll.0.as_raw_fd(), OFlag::O_CLOEXEC).map(|fd| {
             // SAFETY: Used to prevent leaks in sandbox process:
-            std::env::set_var(ENV_IPC_POLL_FD, fd.to_string());
+            let mut fd_str = itoa::Buffer::new();
+            let fd_str = fd_str.format(fd);
+            std::env::set_var(ENV_IPC_POLL_FD, fd_str);
 
             // SAFETY: duprand returns a valid FD on success.
             unsafe { OwnedFd::from_raw_fd(fd) }
@@ -517,37 +547,108 @@ impl IpcWorker {
         // Lookup connection by fd.
         // We need a mutable reference; use get_mut.
         if let Some(conn) = connections.get_mut(&fd) {
-            let mut buf = [0u8; 1024];
             loop {
+                let mut buf = [0u8; LINE_MAX];
+                let mut iov = [IoSliceMut::new(&mut buf)];
+                let mut cmsg = cmsg_space!(UnixCredentials);
+
                 #[allow(clippy::arithmetic_side_effects)]
                 #[allow(unreachable_patterns)]
-                match conn.stream.read(&mut buf).map_err(|e| err2no(&e)) {
-                    Ok(0) => {
+                match recvmsg::<()>(
+                    conn.stream.as_raw_fd(),
+                    &mut iov,
+                    Some(&mut cmsg),
+                    MsgFlags::MSG_DONTWAIT,
+                ) {
+                    Ok(msg) if msg.bytes == 0 => {
                         // EOF on read side: if there's a pending write,
                         // defer closing until after flush.
-                        if let Some(conn) = connections.get_mut(&fd) {
-                            if conn.write_buf.is_empty() {
-                                // no response queued => close immediately.
-                                self.close_connection(connections, fd)?;
-                            } else {
-                                // response pending => mark to close after writing.
-                                conn.should_close = true;
-                            }
+                        if conn.write_buf.is_empty() {
+                            // no response queued => close immediately.
+                            self.close_connection(connections, fd)?;
+                        } else {
+                            // response pending => mark to close after writing.
+                            conn.should_close = true;
                         }
                         return Ok(());
                     }
-                    Ok(n) if conn.buf.len() + n >= LINE_MAX => {
-                        // Input too large -> close conection.
-                        self.close_connection(connections, fd)?;
+                    Ok(msg) if conn.buf.len() + msg.bytes >= LINE_MAX => {
+                        // Input too large:
+                        // 1. Add EPOLLOUT to interests.
+                        // 2. Reject with error message.
+                        // 3. Close connection after reply.
+                        let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
+                        conn.ctl(&epoll, EpollFlags::EPOLLOUT)?;
+                        conn.enqueue_response(IPC_RATE);
+                        conn.should_close = true;
+
                         return Ok(());
                     }
-                    Ok(n) => {
-                        conn.buf.extend_from_slice(&buf[..n]);
+                    Ok(msg) => {
+                        // Authenticate each and every message.
+                        let mut auth = false;
+                        if let Ok(cmsgs) = msg.cmsgs() {
+                            for cmsg in cmsgs {
+                                let creds = if let ControlMessageOwned::ScmCredentials(creds) = cmsg
+                                {
+                                    creds
+                                } else {
+                                    continue;
+                                };
+
+                                if creds.pid() == 0 {
+                                    // Invalid credentials, skip.
+                                    continue;
+                                }
+
+                                if creds.uid() == 0 && creds.gid() == 0 {
+                                    // Matched UID and GID for ROOT: Authenticated.
+                                    auth = true;
+                                    break;
+                                }
+
+                                if let Some(uid) = self.creds.0 {
+                                    if creds.uid() != uid.as_raw() {
+                                        // UID mismatch.
+                                        continue;
+                                    }
+                                }
+
+                                if let Some(gid) = self.creds.1 {
+                                    if creds.gid() != gid.as_raw() {
+                                        // GID mismatch.
+                                        continue;
+                                    }
+                                }
+
+                                // Matched UID and GID: Authenticated.
+                                auth = true;
+                                break;
+                            }
+                        }
+
+                        if !auth {
+                            // Authentication failed:
+                            // 1. Add EPOLLOUT to interests.
+                            // 2. Reject with error message.
+                            // 3. Close connection after reply.
+                            let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
+                            conn.ctl(&epoll, EpollFlags::EPOLLOUT)?;
+                            conn.enqueue_response(IPC_AUTH);
+                            conn.should_close = true;
+
+                            return Ok(());
+                        }
+
+                        // Extend buffer with the message content.
+                        for iov in msg.iovs() {
+                            conn.buf.extend_from_slice(iov);
+                        }
 
                         // Process any full lines in buffer.
                         while let Some(pos) = memchr(b'\n', &conn.buf) {
                             // Handle the line (no allocations needed here).
-                            self.process_command(conn, fd, pos)?;
+                            self.process_command(conn, pos)?;
 
                             // Remove the line *and* the '\n' from the buffer.
                             conn.buf.drain(..=pos);
@@ -576,13 +677,17 @@ impl IpcWorker {
 
         if let Some(conn) = connections.get_mut(&fd) {
             while conn.write_pos < conn.write_buf.len() {
+                let iov = [IoSlice::new(&conn.write_buf[conn.write_pos..])];
+
                 #[allow(clippy::arithmetic_side_effects)]
                 #[allow(unreachable_patterns)]
-                match conn
-                    .stream
-                    .write(&conn.write_buf[conn.write_pos..])
-                    .map_err(|e| err2no(&e))
-                {
+                match sendmsg::<()>(
+                    conn.stream.as_raw_fd(),
+                    &iov,
+                    &[],
+                    MsgFlags::MSG_DONTWAIT,
+                    None,
+                ) {
                     Ok(0) => {
                         // Would block or closed; stop for now.
                         break;
@@ -605,14 +710,9 @@ impl IpcWorker {
 
             if conn.write_pos >= conn.write_buf.len() {
                 // All data written: clear buffer and disable EPOLLOUT.
+                conn.ctl(&epoll, EpollFlags::EPOLLIN)?;
                 conn.write_buf.clear();
                 conn.write_pos = 0;
-                #[allow(clippy::cast_sign_loss)]
-                let event = libc::epoll_event {
-                    events: EpollFlags::EPOLLIN.bits() as u32,
-                    u64: fd as u64,
-                };
-                epoll_ctl_mod_safe(&epoll.0, fd, event)?;
 
                 // If mode is NonInteractive and the last command closed the connection,
                 // we close the connection here.
@@ -626,7 +726,7 @@ impl IpcWorker {
     }
 
     // Process at `pos` boundary from connection `conn` on `fd`.
-    fn process_command(&self, conn: &mut Connection, fd: RawFd, pos: usize) -> Result<(), Errno> {
+    fn process_command(&self, conn: &mut Connection, pos: usize) -> Result<(), Errno> {
         let epoll = self.epoll.as_ref().ok_or(Errno::EADDRNOTAVAIL)?;
 
         // Convert to UTF-8, errors are fatal.
@@ -687,7 +787,7 @@ impl IpcWorker {
                     Cow::Borrowed(IPC_PINK)
                 } else if part == "version" {
                     Cow::Owned(format!(
-            "{{\"major\":{API_VERSION},\"minor\":{IPC_VERSION},\"version\":\"{API_VERSION}.{IPC_VERSION}\"}}\n").into_bytes())
+            "{{\"major\":{API_MAJOR_VERSION},\"minor\":{IPC_MINOR_VERSION},\"version\":\"{API_MAJOR_VERSION}.{IPC_MINOR_VERSION}\"}}\n").into_bytes())
                 } else {
                     // Regular command: pass to handle_cmd.
                     Cow::Owned(self.handle_cmd(&cmd)?)
@@ -709,12 +809,7 @@ impl IpcWorker {
 
         // After enqueueing data, ensure EPOLLOUT is enabled.
         if !conn.write_buf.is_empty() {
-            #[allow(clippy::cast_sign_loss)]
-            let event = libc::epoll_event {
-                events: (EpollFlags::EPOLLIN | EpollFlags::EPOLLOUT).bits() as u32,
-                u64: fd as u64,
-            };
-            epoll_ctl_mod_safe(&epoll.0, fd, event)?;
+            conn.ctl(&epoll, EpollFlags::EPOLLIN | EpollFlags::EPOLLOUT)?;
         }
 
         Ok(())
